import os
import sys
from time import sleep

import numpy
from setuptools import Extension
from setuptools.command.build_ext import build_ext
from setuptools.errors import PlatformError, CCompilerError, ExecError, PlatformError

USE_CYTHON = os.environ.get('USE_CYTHON', False)
CYTHON = False
try:
    from Cython.Build import cythonize

    CYTHON = True
except ImportError:
    USE_CYTHON = False

file_ext = '.pyx' if USE_CYTHON else '.c'

build_modules = ['convertor']

extensions = [
    Extension(f'dfpwm.{module}', [f'./dfpwm/{module}{file_ext}'])
    for module in build_modules
]

include_dirs = [
    numpy.get_include()
]


class ExtBuilder(build_ext):
    def run(self):
        try:
            super().run()
        except (PlatformError, FileNotFoundError):
            raise Exception('File not found. Could not compile C extension.')

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except (CCompilerError, ExecError, PlatformError, ValueError):
            raise Exception('Could not compile C extension.')


def build(setup_kwargs: dict):
    os.environ['CFLAGS'] = '-O3'

    setup_kwargs.update({
        'ext_modules': cythonize(
            extensions, language_level=3,
            compiler_directives={'linetrace': True}
        ) if USE_CYTHON else extensions, "include_dirs": include_dirs,
        "cmdclass": {
            "build_ext": ExtBuilder,
        }
    })

    if __name__ == '__main__':
        if len(sys.argv) != 0 and sys.argv[0] == 'build':
            if not CYTHON:
                print("Please install cython to convert to .c file")
                sys.exit(-1)

            cythonize([f'./dfpwm/{module}.pyx' for module in build_modules])



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
import typing
if typing.TYPE_CHECKING:
    from typing import Callable, Iterable

from libc.math cimport floor
from libc.stdlib cimport malloc, free
import numpy as np
cimport numpy as np

cdef int SAMPLE_RATE = 48000
cdef int PREC = 10

cdef short* proc_untracked(const float[:] input_data, const int sz) nogil:
    cdef int charge = 0
    cdef int strength = 0
    cdef short previous_bit = False

    cdef short *out = <short*>malloc(sizeof(short) * sz)

    cdef size_t i = 0
    cdef size_t j = 0

    cdef int this_byte = 0
    cdef float level
    cdef short current_bit
    cdef int next_charge
    cdef int next_strength
    cdef int target
    cdef int v1, v2
    while i < sz:
        j = 0
        while j < 8:
            level = floor(input_data[i * 8 + j] * 127)

            current_bit = level > charge or (level == charge and charge == 127)
            target = (127 if current_bit else -128)

            v1 = ((strength * (target - charge) + (1 << PREC - 1)) >> PREC)

            next_charge = charge + v1
            if v1 == 0 and next_charge != target:
                next_charge += 1 if current_bit else -1

            v2 = this_byte >> 1

            this_byte = v2 + 128 if current_bit else v2

            z = ((1 << PREC) - 1 if current_bit == previous_bit else 0)
            next_strength = strength

            if strength != z:
                next_strength += 1 if current_bit == previous_bit else -1

            v3 = 2 << (PREC - 8)
            if next_strength < v3:
                next_strength = v3

            charge = next_charge
            strength = next_strength
            previous_bit = current_bit

            j+=1

        out[i] = this_byte

        i += 1

    return out

cpdef compressor(
        in_array,
        tracker: Callable[[Iterable], Iterable] = None
):
    cdef int charge = 0
    cdef int strength = 0
    cdef char previous_bit = False

    cdef size_t sz = len(in_array) // 8
    cdef const float [:] in_array_view = in_array

    cdef short *out_array = <short*>malloc(sizeof(short) * sz)

    cdef size_t i
    cdef size_t j
    cdef float level
    cdef short this_byte
    cdef char current_bit
    cdef int next_charge
    cdef int next_strength
    cdef int target
    cdef int v1, v2

    if tracker:
        for i in tracker(range(sz)):
            this_byte = 0

            for j in range(8):
                level = floor(in_array_view[i * 8 + j] * 127)

                current_bit = level > charge or (level == charge and charge == 127)
                target = (127 if current_bit else -128)

                v1 = ((strength * (target - charge) + (1 << PREC - 1)) >> PREC)

                next_charge = charge + v1
                if v1 == 0 and next_charge != target:
                    next_charge += 1 if current_bit else -1

                v2 = this_byte >> 1

                this_byte = v2 + 128 if current_bit else v2

                z = ((1 << PREC) - 1 if current_bit == previous_bit else 0)
                next_strength = strength

                if strength != z:
                    next_strength += 1 if current_bit == previous_bit else -1

                v3 = 2 << (PREC - 8)
                if next_strength < v3:
                    next_strength = v3

                charge = next_charge
                strength = next_strength
                previous_bit = current_bit

            out_array[i] = this_byte
    else:
        free(<void*>out_array)
        array = proc_untracked(in_array_view, sz)
        out_array = array

    cdef np.ndarray result = np.asarray(<short[:sz]>out_array).astype('int8')
    free(<void*>array)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dfpwm']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.26.2,<2.0.0']

setup_kwargs = {
    'name': 'dfpwm',
    'version': '0.1.4.1',
    'description': 'DFPWM convertor for Python',
    'long_description': '# DFPWM\n\nDFPWM convertor for Python\n\n## Usage\n\n```python\nfrom pathlib import Path\nimport soundfile as sf  # for reading audio\nimport dfpwm\n\ndata, sample_rate = sf.read(\'./someaudio.mp3\')  # read audio\n\n# If sample rate is not 48000, may get strange result\n# use `dfpwm.resample(...)` to resample\nif sample_rate != dfpwm.SAMPLE_RATE:\n    raise ValueError(f"{sample_rate} != {dfpwm.SAMPLE_RATE}")\n\nif len(data.shape) != 0 and data.shape[1] > 1:\n    data = data[:, 0]  # get channel 0\n\ndfpwm = dfpwm.compressor(data)  # convert\nPath(\'out.dfpwm\').write_bytes(dfpwm)  # write result to file\n```\n\n## Build from source\n\n### Clone\n\n```shell\ngit clone https://github.com/CyanChanges/python-dfpwm.git python-dfpwm\ncd python-dfpwm\n```\n\n### Build\nThis project use `poetry` to build,\nMake sure `poetry` is installed.\n\n```shell\npoetry build -f sdist\n```\n\n',
    'author': 'cyan',
    'author_email': 'contact@cyans.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from typing import BinaryIO

import numpy as np

SAMPLE_RATE = 48000


def resample_from_file(io: "BinaryIO") -> tuple["np.ndarray", float]:
    import warnings
    import librosa
    warnings.warn("resample_from_file is deprecated, use dfpwm.resample instead")
    return librosa.load(io, dtype='float32', sr=SAMPLE_RATE)


def resample(data: np.ndarray, origin_samplerate: float, target_sample_rate=SAMPLE_RATE):
    import librosa
    return librosa.resample(data, origin_samplerate, target_sample_rate)
